import ctypes
import time
from numbers import Integral
from typing import Any, Optional, Sequence

import numpy as np
from sklearn.base import (
    BaseEstimator,
    ClassNamePrefixFeaturesOutMixin,
    ClusterMixin,
    _fit_context,
)
from sklearn.exceptions import NotFittedError
from sklearn.utils._param_validation import Interval

import bico._core  # type: ignore

_DLL = ctypes.cdll.LoadLibrary(bico._core.__file__)


class BICO(BaseEstimator, ClusterMixin, ClassNamePrefixFeaturesOutMixin):
    _parameter_constraints: dict = {
        "n_clusters": [Interval(Integral, 1, None, closed="left")],
        "summary_size": [None, Interval(Integral, 1, None, closed="left")],
        "n_random_projections": [None, Interval(Integral, 1, None, closed="left")],
        "random_state": [None, Interval(Integral, 0, None, closed="left")],
        "fit_coreset": [bool],
        "coreset_estimator": [None, BaseEstimator],
    }

    _CORESET_ESTIMATOR_ERROR = (
        "BICO computes a coreset. If you want to compute labels, "
        "then you need to call another estimator on the coreset. "
        "This is done automatically by using the `fit_coreset=True` parameter during "
        "the initialization of the class. "
        "The default estimator is KMeans, but you can pass any other estimator to "
        "the `coreset_estimator` parameter."
    )

    _NOT_FITTED_CORESET_ERROR = (
        "This BICO instance is not fitted yet. Call `fit` "
        "or `partial_fit` before using its results. If you are using "
        "`partial_fit`, then you need to call `partial_fit` with no arguments to "
        "compute the coreset."
    )

    def __init__(
        self,
        n_clusters: int,
        summary_size: Optional[int] = None,
        n_random_projections: Optional[int] = None,
        random_state: Optional[int] = None,
        fit_coreset: bool = False,
        coreset_estimator: Optional[BaseEstimator] = None,
    ):
        self.n_clusters = n_clusters
        self.summary_size = (
            summary_size if summary_size is not None else n_clusters * 200
        )
        self.n_random_projections = n_random_projections
        self.random_state = random_state
        self.fit_coreset = fit_coreset
        self.coreset_estimator = coreset_estimator

    @property
    def labels_(self) -> np.ndarray:
        if not hasattr(self, "_labels"):
            raise NotFittedError(self._CORESET_ESTIMATOR_ERROR)
        return self._labels

    @property
    def inertia_(self) -> float:
        if not hasattr(self, "_inertia"):
            raise NotFittedError(self._CORESET_ESTIMATOR_ERROR)
        return self._inertia

    @property
    def cluster_centers_(self) -> np.ndarray:
        if not hasattr(self, "_cluster_centers"):
            raise NotFittedError(self._CORESET_ESTIMATOR_ERROR)
        return self._cluster_centers

    @property
    def coreset_points_(self) -> np.ndarray:
        if not hasattr(self, "_coreset_points"):
            raise NotFittedError(self._NOT_FITTED_CORESET_ERROR)
        return self._coreset_points

    @property
    def coreset_weights_(self) -> np.ndarray:
        if not hasattr(self, "_coreset_weights"):
            raise NotFittedError(self._NOT_FITTED_CORESET_ERROR)
        return self._coreset_weights

    @_fit_context(prefer_skip_nested_validation=True)
    def fit(
        self,
        X: Sequence[Sequence[float]],
        y: Any = None,
    ) -> "BICO":
        return self._fit(X, partial=False)

    @_fit_context(prefer_skip_nested_validation=True)
    def partial_fit(
        self,
        X: Optional[Sequence[Sequence[float]]] = None,
        y: Any = None,
    ) -> "BICO":
        if X is None:
            self._compute_coreset()
            return self
        else:
            return self._fit(X, partial=True)

    def _fit_coreset(
        self,
    ) -> None:
        if self.coreset_estimator is None:
            from sklearn.cluster import KMeans

            self.coreset_estimator = KMeans(
                n_clusters=self.n_clusters,
                random_state=self.random_state,
            )

        self.coreset_estimator.fit(
            self._coreset_points, sample_weight=self._coreset_weights
        )
        self._cluster_centers: np.ndarray = self.coreset_estimator.cluster_centers_
        self._labels: np.ndarray = self.coreset_estimator.labels_
        self._inertia: float = self.coreset_estimator.inertia_

    def _compute_coreset(self, fit_coreset: bool = False) -> "BICO":
        if not hasattr(self, "bico_obj_"):
            raise NotFittedError(
                "This BICO instance is not fitted yet. " "Call `fit` or `partial_fit`."
            )

        c_coreset_weights = (ctypes.c_double * self.summary_size)()
        c_points = (ctypes.c_double * self.n_features_in_ * self.summary_size)()

        _DLL.compute.restype = ctypes.c_size_t
        n_found_points = _DLL.compute(self.bico_obj_, c_coreset_weights, c_points)

        self._coreset_weights: np.ndarray = np.ctypeslib.as_array(c_coreset_weights)[
            :n_found_points
        ]
        self._coreset_points: np.ndarray = np.ctypeslib.as_array(
            c_points, shape=(self.summary_size, self.n_features_in_)
        )[:n_found_points]

        self._n_features_out = n_found_points

        if self.fit_coreset or fit_coreset:
            self._fit_coreset()

        return self

    def _fit(
        self, X: Sequence[Sequence[float]], partial: bool, fit_coreset: bool = False
    ) -> "BICO":
        has_bico_obj = getattr(self, "bico_obj_", None)
        first_call = not (partial and has_bico_obj)

        _X = np.array(X, dtype=np.float64, order="C", copy=False)
        assert isinstance(_X, np.ndarray)

        if first_call:
            self.n_features_in_: int = _X.shape[1]
            _seed = int(time.time()) if self.random_state is None else self.random_state

            # In Melanie's thesis, p = d
            if self.n_random_projections is None:
                self.n_random_projections = self.n_features_in_

            # Declare c types
            c_d = ctypes.c_uint(self.n_features_in_)
            c_k = ctypes.c_uint(self.n_clusters)
            c_p = ctypes.c_uint(self.n_random_projections)
            c_m = ctypes.c_uint(self.summary_size)
            c_random_state = ctypes.c_size_t(_seed)

            _DLL.init.restype = ctypes.POINTER(ctypes.c_void_p)
            self.bico_obj_: Any = _DLL.init(c_d, c_k, c_p, c_m, c_random_state)

        c_array = _X.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
        c_n = ctypes.c_uint(_X.shape[0])
        _DLL.addData(self.bico_obj_, c_array, c_n)

        if not partial or fit_coreset:
            self._compute_coreset(fit_coreset)

        return self

    def __del__(self) -> None:
        if hasattr(self, "bico_obj_"):
            _DLL.freeBico(self.bico_obj_)
            del self.bico_obj_

    def fit_predict(
        self, X: Sequence[Sequence[float]], y: Any = None, **kwargs: Any
    ) -> np.ndarray:
        self._fit(X, partial=False, fit_coreset=True)
        return self.labels_

    def predict(self, X: Sequence[Sequence[float]]) -> Any:
        self._fit_coreset()

        if self.coreset_estimator is None:
            raise NotFittedError(self._CORESET_ESTIMATOR_ERROR)

        return self.coreset_estimator.predict(X)



from typing import Any, Dict

from setuptools import Extension
from setuptools.command.build_ext import build_ext
from setuptools.errors import CompileError

extension = Extension(
    name="bico._core",
    sources=[
        "bico/_core.cpp",
        "bico/point/pointcentroid.cpp",
        "bico/point/squaredl2metric.cpp",
        "bico/point/point.cpp",
        "bico/point/realspaceprovider.cpp",
        "bico/point/l2metric.cpp",
    ],
    include_dirs=["bico"],
)

# Thank you https://github.com/dstein64/kmeans1d!


class BuildExt(build_ext):
    """A custom build extension for adding -stdlib arguments for clang++."""

    def build_extensions(self) -> None:
        # '-std=c++11' is added to `extra_compile_args` so the code can compile
        # with clang++. This works across compilers (ignored by MSVC).
        for extension in self.extensions:
            extension.extra_compile_args.append("-std=c++11")
        try:
            build_ext.build_extensions(self)
        except CompileError:
            # Workaround Issue #2.
            # '-stdlib=libc++' is added to `extra_compile_args` and `extra_link_args`
            # so the code can compile on macOS with Anaconda.
            for extension in self.extensions:
                extension.extra_compile_args.append("-stdlib=libc++")
                extension.extra_link_args.append("-stdlib=libc++")
            build_ext.build_extensions(self)


def build(setup_kwargs: Dict[str, Any]) -> None:
    setup_kwargs.update(
        {"ext_modules": [extension], "cmdclass": {"build_ext": BuildExt}}
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bico']

package_data = \
{'': ['*'],
 'bico': ['base/*',
          'clustering/*',
          'datastructure/*',
          'evaluation/*',
          'exception/*',
          'point/*']}

install_requires = \
['numpy>=1.26.4,<2.0.0', 'scikit-learn>=1.5.0,<2.0.0']

setup_kwargs = {
    'name': 'bico',
    'version': '0.1.1',
    'description': 'BICO is a fast streaming algorithm to compute coresets for the k-means problem on very large sets of points.',
    'long_description': '[![Build Status](https://github.com/algo-hhu/bico/actions/workflows/mypy-flake-test.yml/badge.svg)](https://github.com/algo-hhu/bico/actions)\n[![License: GPL v3](https://img.shields.io/badge/License-GPLv3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0)\n[![Supported Python version](https://img.shields.io/badge/python-3.9+-blue.svg)](https://www.python.org/downloads/release/python-390/)\n[![Stable Version](https://img.shields.io/pypi/v/bico?label=stable)](https://pypi.org/project/bico/)\n\n# BICO\n\nBICO is a fast streaming algorithm to compute high quality solutions for the k-means problem on very large sets of points. It combines the tree data structure of SIGMOND Test of Time Award winning algorithm BIRCH with insights from clustering theory to obtain solutions fast while keeping the error regarding the k-means cost function low.\n\n<!---\nTODO: Add logo\n<p align="center">\n  <img src="https://raw.githubusercontent.com/algo-hhu/bico/main/images/logo.png" alt="BICO Logo"/>\n</p>\n-->\n\n## Installation\n\n```bash\npip install bico\n```\n\n## Example\n\n```python\nfrom bico import BICO\nimport numpy as np\nimport time\n\nnp.random.seed(42)\n\ndata = np.random.rand(10000, 10)\n\nstart = time.time()\nbico = BICO(n_clusters=3, random_state=0, fit_coreset=True)\nbico.fit(data)\n\nprint("Time:", time.time() - start)\n# Time: 0.08275651931762695\n\nprint(bico.coreset_points_)\n# BICO returns a set of points that act as a summary of the entire dataset.\n# By default, at most 200 * n_clusters points are returned.\n# This behaviour can be changed by setting the `summary_size` parameter.\n\n# [[0.45224018 0.70183673 0.55506671 ... 0.70132665 0.57244196 0.66789088]\n#  [0.73712952 0.5250208  0.43809322 ... 0.61427161 0.67910981 0.56207661]\n#  [0.89905336 0.46942062 0.20677639 ... 0.74210482 0.75714522 0.49651055]\n#  ...\n#  [0.68744494 0.41508081 0.39197623 ... 0.44093386 0.21983902 0.37237243]\n#  [0.60820965 0.29406341 0.67067782 ... 0.66435474 0.2390822  0.20070476]\n#  [0.67385626 0.33474823 0.68238779 ... 0.3581703  0.65646253 0.41386131]]\n\nprint(bico.cluster_centers_)\n# If the `fit_coreset` parameter is set to True, the cluster centers are computed using KMeans from sklearn based on the coreset.\n\n# [[0.46892639 0.41968333 0.47302945 0.51782955 0.39390839 0.56209413\n#   0.4481691  0.49521457 0.31394509 0.5104331 ]\n#  [0.54384638 0.518978   0.49456809 0.56677848 0.63881783 0.33627504\n#   0.49873782 0.5541338  0.52913562 0.56017203]\n#  [0.48639347 0.55542596 0.54350474 0.41931257 0.48117255 0.60089563\n#   0.55457724 0.44833238 0.67583389 0.43069267]]\n```\n\n## Example with Large Datasets\n\nFor very large datasets, the data may not actually fit in memory. In this case, you can use `partial_fit` to stream the data in chunks. In this example, we use the [US Census Data (1990) dataset](https://archive.ics.uci.edu/dataset/116/us+census+data+1990). You can find more examples in the [tests](./tests/test.py) folder.\n\n```python\nfrom bico import BICO\nimport numpy as np\nimport time\n\nnp.random.seed(42)\n\ndata = np.random.rand(10000, 10)\n\nstart = time.time()\nbico = BICO(n_clusters=3, random_state=0)\nfor chunk in pd.read_csv(\n    "census.txt", delimiter=",", header=None, chunksize=10000\n):\n    bico.partial_fit(chunk.to_numpy(copy=False))\n# If a final `partial_fit` is called with no data, the coreset is computed\nbico.partial_fit()\n```\n\n## Development\n\nInstall [poetry](https://python-poetry.org/docs/#installation)\n```bash\ncurl -sSL https://install.python-poetry.org | python3 -\n```\n\nInstall clang\n```bash\nsudo apt-get install clang\n```\n\nSet clang variables\n```bash\nexport CXX=/usr/bin/clang++\nexport CC=/usr/bin/clang\n```\n\nInstall the package\n```bash\npoetry install\n```\n\nIf the installation does not work and you do not see the C++ output, you can build the package to see the stack trace\n```bash\npoetry build\n```\n\nRun the tests\n```bash\npoetry run python -m unittest discover tests -v\n```\n\n## Citation\n\nIf you use this code, please cite [the following paper](https://link.springer.com/chapter/10.1007/978-3-642-40450-4_41):\n\n```\nHendrik Fichtenberger, Marc Gillé, Melanie Schmidt, Chris Schwiegelshohn and Christian Sohler. "BICO: BIRCH Meets Coresets for k-Means Clustering" (2013). ESA 2013.\n```\n',
    'author': 'Melanie Schmidt',
    'author_email': 'mschmidt@hhu.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}
from build_extension import *
build(setup_kwargs)

setup(**setup_kwargs)
